#include "config.h"

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "skel.h"
#include "skin.h"
#include "real.h"
#include "vec/vec.h"
#include "vec/mtx.h"
#include "vec/qtr.h"

static float ctbl[256];
static float stbl[256];

struct SktrTag * skel_init(const char * name)
{
  int i;
  double delta = 360.0 / 256.0;
  double angle = 0.0;
  
  struct SkelTag * skel = skl_init(stream_init(name));
  struct SktrTag * sktr;

  /* 
   * not having constructors (non constant initializers) in C
   * sucks big time
   */
  for (i = 0; i < 256; i++, angle += delta)
  {
    ctbl[i] = (float) cos(DEG2RAD * angle);
    stbl[i] = (float) sin(DEG2RAD * angle);
  }

  #define VECSZ VECSIZE * sizeof(float) * skel->LimbCount
  #define QTRSZ QTRSIZE * sizeof(float) * skel->LimbCount

  sktr = malloc(SKTRTAGSZ);
  _CHECK(!sktr, "malloc(%u) failed", SKTRTAGSZ, 0);

  sktr->skel = skel;

  sktr->LimbListS          = malloc(VECSZ);
  sktr->LimbListE          = malloc(VECSZ);
  sktr->AQuaternionList    = malloc(QTRSZ);
  sktr->BQuaternionList    = malloc(QTRSZ);
  sktr->TempQuaternionList = malloc(QTRSZ);
  sktr->SkelQuaternionList = malloc(QTRSZ);
  sktr->ItrpQuaternionList = malloc(QTRSZ);

  _CHECK(!sktr->LimbListS,          "malloc(%u) failed", VECSZ, 0);
  _CHECK(!sktr->LimbListE,          "malloc(%u) failed", VECSZ, 0);
  _CHECK(!sktr->TempQuaternionList, "malloc(%u) failed", QTRSZ, 0);
  _CHECK(!sktr->SkelQuaternionList, "malloc(%u) failed", QTRSZ, 0);
  _CHECK(!sktr->ItrpQuaternionList, "malloc(%u) failed", QTRSZ, 0);
  _CHECK(!sktr->AQuaternionList,    "malloc(%u) failed", QTRSZ, 0);
  _CHECK(!sktr->BQuaternionList,    "malloc(%u) failed", QTRSZ, 0);

  memset(sktr->AQuaternionList, 0,  QTRSZ);
  memset(sktr->BQuaternionList, 0,  QTRSZ);

  #undef VECSZ
  #undef QTRSZ
  
  sktr->SlerpStep      = 0;
  sktr->SlerpStepCount = 0;

  return sktr;
}

void skel_done(struct SktrTag * sktr)
{
  #define _FREE(p) if (p) free(p)

  _FREE(sktr->LimbListS);
  _FREE(sktr->LimbListE);
  _FREE(sktr->TempQuaternionList);
  _FREE(sktr->ItrpQuaternionList);
  _FREE(sktr->SkelQuaternionList);
  _FREE(sktr->AQuaternionList);
  _FREE(sktr->BQuaternionList);
  _FREE(sktr->CosList);
  _FREE(sktr->SinList);
  _FREE(sktr->ULenList);
  _FREE(sktr->VLenList);
  if (sktr->skel)
  {
    _FREE(sktr->skel->LimbList);
    free(sktr->skel);
  }
  free(sktr);
}

static void rotation_YXZ(float * R, uchar xa, uchar ya, uchar za)
{
  float Rxy[4], Rx[4], Ry[4], Rz[4];

  #define ANGLE(a) ((a) * 360.f / 256.f)
  float ax = ANGLE(xa);
  float ay = ANGLE(ya);
  float az = ANGLE(za);
  #undef ANGLE

  qtr_rotation(Rx, ax, 1.f, 0.f, 0.f);
  qtr_rotation(Ry, ay, 0.f, 1.f, 0.f);
  qtr_rotation(Rz, az, 0.f, 0.f, 1.f);
  qtr_qtrmul(Rxy, Rx, Ry);
  qtr_qtrmul(R, Rz, Rxy);
}

#define LIMBS(i) (&LimbListS[(i) * VECSIZE])
#define LIMBE(i) (&LimbListE[(i) * VECSIZE])
#define TEMPQTR(i) (&sktr->TempQuaternionList[(i) * QTRSIZE])
#define SKELQTR(i) (&sktr->SkelQuaternionList[(i) * QTRSIZE])

void skel_process(struct SktrTag * sktr,
                  struct RetrTag * retr,
                  struct SntrTag * sntr)
{
  int              i;
  int              SkinNo;
  float          * LimbListS;
  float          * LimbListE;
  float          * SkelQtr;
  float          * TempQtr;
  struct SkelTag * skel;
  struct LimbTag * limb;

  _CHECK(!sntr->VertLimbList, "skin_process probably wasnt called", 0, 0);

  #define _SZ sizeof(float) * retr->real->VertexCount
  sktr->SinList  = malloc(_SZ);
  sktr->CosList  = malloc(_SZ);
  sktr->ULenList = malloc(_SZ);
  sktr->VLenList = malloc(_SZ);
  
  _CHECK(!sktr->SinList,  "malloc(%u) failed", _SZ, 0);
  _CHECK(!sktr->CosList,  "malloc(%u) failed", _SZ, 0);
  _CHECK(!sktr->ULenList, "malloc(%u) failed", _SZ, 0);
  _CHECK(!sktr->VLenList, "malloc(%u) failed", _SZ, 0);
  #undef _SZ

  SkinNo             = 0;
  skel               = sktr->skel;
  limb               = skel->LimbList;
  SkelQtr            = sktr->SkelQuaternionList;
  TempQtr            = sktr->TempQuaternionList;
  LimbListS          = sktr->LimbListS;
  LimbListE          = sktr->LimbListE;

  for (i = 0; i < skel->LimbCount; i++, limb++, SkelQtr += 4, TempQtr += 4)
  {
    int parent = limb->parent;
    uchar xa  = limb->xa;
    uchar ya  = limb->ya;
    uchar za  = limb->za;
    float I[3], J[3], K[3];

/*    printf("%d: %s %p %p\n", i, limb->name, SkelQtr, TempQtr);*/

    if (parent == -1)
    {
      /*
       * not exactly true but heck i'm tired of matrix math
       */
      QTRIDN(SkelQtr);
      QTRIDN(TempQtr);
      VECNIL(LIMBS(i));
      VECNIL(LIMBE(i));
      VECPAK(I, 0.f, 1.f, 0.f);
      VECPAK(J, 1.f, 0.f, 0.f);
      VECPAK(K, 0.f, 0.f, 1.f);
    } else
    {
      float Ryxz[4];

      /*
       * coordinate system in kindof screen system
       * OY - going inside
       * OX - going left 
       * OZ - going up
       * ya rotation around OY (can be thought as rotation around limb)
       * xa rotation around OX
       * za rotation around OZ
       * (i have heard that the same system is used in 3-D Studio)
       */ 

      rotation_YXZ(Ryxz, xa, ya, za);

      QTRCPY(SkelQtr, Ryxz);
      qtr_qtrmul(TempQtr, TEMPQTR(parent), Ryxz);

      qtr_vecmul(I, TempQtr, 0.f, 1.f, 0.f);
      qtr_vecmul(J, TempQtr, 1.f, 0.f, 0.f);
      qtr_vecmul(K, TempQtr, 0.f, 0.f, 1.f);

      VECCPY(LIMBS(i), LIMBE(parent));
      {
        float D[3];
        long dx = limb->dx;
        long dy = limb->dy;
        long dz = limb->dz;
        if (dx || dy || dz)
        {
          qtr_vecmul(D, TEMPQTR(parent), (float) dx, (float) dy, (float) dz);
          VECADD(LIMBS(i), D);
        }
      }

      VECCPY(LIMBE(i), I);
      VECSCL(LIMBE(i), limb->length);
      VECADD(LIMBE(i), LIMBS(i));
      #if 0
      VECPRT(LIMBS(i));
      VECPRT(LIMBE(i));
      #endif
    }

    /*
     * convert vertex coordinates from affine to limb relative(cylindric) 
     * system
     */

    while (sntr->VertLimbList[SkinNo * 2] == i)
    {
      int VertexNo = sntr->VertLimbList[SkinNo * 2 + 1];

      long x = retr->real->VertexList[VertexNo].x;
      long y = retr->real->VertexList[VertexNo].y;
      long z = retr->real->VertexList[VertexNo].z;

      float OV[3];
      float VS[3];
      float ES[3];
      float ulen, vlen;
      float cos_j, sin_j;

      _CHECK(VertexNo > retr->real->VertexCount,
             "vertex overflow %d > %d", VertexNo,
             retr->real->VertexCount);

      VECPAK(OV, x, y, z);

      VECCPY(VS, OV);
      VECSUB(VS, LIMBS(i));
      ulen = VECDOT(VS, I);

      VECCPY(ES, I);
      VECSCL(ES, ulen);
      VECADD(ES, LIMBS(i));
      VECSUB(ES, OV);

      vlen = VECMAG(ES);
      VECDIV(ES, vlen);

      cos_j = -(VECDOT(ES, J));
      sin_j = VECDOT(ES, K);

      sktr->CosList[VertexNo] = cos_j;
      sktr->SinList[VertexNo] = sin_j;
      sktr->ULenList[VertexNo] = ulen;
      sktr->VLenList[VertexNo] = vlen;

      /*
       * for shape to render properly without calling skel_apply or skel_next
       *
       */
      retr->VertexList[VertexNo * VECSIZE + 0] = (float) retr->real->VertexList[VertexNo].x;
      retr->VertexList[VertexNo * VECSIZE + 1] = (float) retr->real->VertexList[VertexNo].y;
      retr->VertexList[VertexNo * VECSIZE + 2] = (float) retr->real->VertexList[VertexNo].z;

      SkinNo++;
    }
  }
}

void skel_start(struct SktrTag * sktr)
{
  int              i;
  float          * BQtr;
  struct LimbTag * limb;

  limb = sktr->skel->LimbList;
  BQtr = sktr->BQuaternionList;

  for (i = 0; i < sktr->skel->LimbCount; i++, limb++, BQtr += 4)
  {
    rotation_YXZ(BQtr, limb->xaa, limb->yaa, limb->zaa);
  }
}

void skel_next(struct SktrTag * sktr,
               struct RetrTag * retr,
               struct SntrTag * sntr)
{
  int              i;
  int              SkinNo;
  float          * LimbListS;
  float          * LimbListE;
  float          * ItrpQtr;
  float          * SkelQtr;
  float          * TempQtr;
  struct SkelTag * skel;
  struct LimbTag * limb;

  SkinNo = 0;
  skel = sktr->skel;

  ItrpQtr   = sktr->ItrpQuaternionList;
  SkelQtr   = sktr->SkelQuaternionList;
  TempQtr   = sktr->TempQuaternionList;
  LimbListS = sktr->LimbListS;
  LimbListE = sktr->LimbListE;

  for (i = 0, limb = skel->LimbList; i < skel->LimbCount; i++, limb++,
                                                          ItrpQtr += 4,
                                                          SkelQtr += 4,
                                                          TempQtr += 4)
  {
    int parent = limb->parent;
    float I[3], J[3];

    if (parent == -1)
    {
      qtr_qtrmul(TempQtr, SkelQtr, ItrpQtr);
      VECNIL(LIMBS(i));
      VECNIL(LIMBE(i));
      qtr_vecmul(I, TempQtr, 0.f, 1.f, 0.f);
      qtr_vecmul(J, TempQtr, 1.f, 0.f, 0.f);
    } else
    {
      float R[4];

      qtr_qtrmul(R, SkelQtr, ItrpQtr);
      qtr_qtrmul(TempQtr, TEMPQTR(parent), R);

      qtr_vecmul(I, TempQtr, 0.f, 1.f, 0.f);
      qtr_vecmul(J, TempQtr, 1.f, 0.f, 0.f);

      VECCPY(LIMBS(i), LIMBE(parent));
      {
        float D[3];
        long dx = limb->dx;
        long dy = limb->dy;
        long dz = limb->dz;
        if (dx || dy || dz)
        {
          qtr_vecmul(D, TEMPQTR(parent), (float) dx, (float) dy, (float) dz);
          VECADD(LIMBS(i), D);
        }
      }

      VECCPY(LIMBE(i), I);
      VECSCL(LIMBE(i), limb->length);
      VECADD(LIMBE(i), LIMBS(i));
    }

    /*
     *  convert vertexes given in cylindrical system (limb based)
     *  based on stored p, z, fi 
     *  and computed i, j
     */

    while (sntr->VertLimbList[SkinNo * 2] == i)
    {
      int VertexNo = sntr->VertLimbList[SkinNo * 2 + 1];
      float R[9];
      float VD[3];
      float UD[3];
      float cos_j, sin_j;
      float ulen, vlen;
        
      cos_j = sktr->CosList[VertexNo];
      sin_j = sktr->SinList[VertexNo];
      ulen = sktr->ULenList[VertexNo];
      vlen = sktr->VLenList[VertexNo];

      mtx_rotationcs(R, cos_j, sin_j, VECUNP(I));

      VECCPY(UD, I);
      VECSCL(UD, ulen);

      mtx_vecmul(VD, R, VECUNP(J));
      VECSCL(VD, vlen);

      retr->VertexList[VertexNo * VECSIZE + 0] = VD[0] + UD[0] + LIMBS(i)[0];
      retr->VertexList[VertexNo * VECSIZE + 1] = VD[1] + UD[1] + LIMBS(i)[1];
      retr->VertexList[VertexNo * VECSIZE + 2] = VD[2] + UD[2] + LIMBS(i)[2];

      SkinNo++;
    }
  }
}

void skel_apply(struct SktrTag * sktr,
                struct RetrTag * retr,
                struct SntrTag * sntr)
{
  int              i;
  int              SkinNo;
  float          * LimbListS;
  float          * LimbListE;
  float          * TempQuaternionList;
  float          * SkelQuaternionList;
  struct SkelTag * skel;
  struct LimbTag * limb;

  SkinNo = 0;
  skel = sktr->skel;

  LimbListS          = sktr->LimbListS;
  LimbListE          = sktr->LimbListE;
  TempQuaternionList = sktr->TempQuaternionList;
  SkelQuaternionList = sktr->SkelQuaternionList;

  for (i = 0, limb = skel->LimbList; i < skel->LimbCount; i++, limb++)
  {
    float Ryxz[4];
    int parent = limb->parent;
    uchar xa = limb->xaa;
    uchar ya = limb->yaa;
    uchar za = limb->zaa;
    float I[3], J[3];

    rotation_YXZ(Ryxz, xa, ya, za);
    if (parent == -1)
    {
      qtr_qtrmul(TEMPQTR(i), SKELQTR(i), Ryxz);
      VECNIL(LIMBS(i));
      VECNIL(LIMBE(i));
      qtr_vecmul(I, TEMPQTR(i), 0.f, 1.f, 0.f);
      qtr_vecmul(J, TEMPQTR(i), 1.f, 0.f, 0.f);
    } else
    {
      float R[4];

      qtr_qtrmul(R, SKELQTR(i), Ryxz);
      qtr_qtrmul(TEMPQTR(i), TEMPQTR(parent), R);

      qtr_vecmul(I, TEMPQTR(i), 0.f, 1.f, 0.f);
      qtr_vecmul(J, TEMPQTR(i), 1.f, 0.f, 0.f);

      VECCPY(LIMBS(i), LIMBE(parent));
      {
        float D[3];
        long dx = limb->dx;
        long dy = limb->dy;
        long dz = limb->dz;
        if (dx || dy || dz)
        {
          qtr_vecmul(D, TEMPQTR(parent), (float) dx, (float) dy, (float) dz);
          VECADD(LIMBS(i), D);
        }
      }

      VECCPY(LIMBE(i), I);
      VECSCL(LIMBE(i), limb->length);
      VECADD(LIMBE(i), LIMBS(i));
    }

    /*
     *  convert vertexes given in cylindrical system (limb based)
     *  based on stored p, z, fi 
     *  and computed i, j
     */

    while (sntr->VertLimbList[SkinNo * 2] == i)
    {
      int VertexNo = sntr->VertLimbList[SkinNo * 2 + 1];
      float R[9];
      float VD[3];
      float UD[3];
      float cos_j, sin_j;
      float ulen, vlen;
        
      cos_j = sktr->CosList[VertexNo];
      sin_j = sktr->SinList[VertexNo];
      ulen = sktr->ULenList[VertexNo];
      vlen = sktr->VLenList[VertexNo];

      mtx_rotationcs(R, cos_j, sin_j, VECUNP(I));

      VECCPY(UD, I);
      VECSCL(UD, ulen);

      mtx_vecmul(VD, R, VECUNP(J));
      VECSCL(VD, vlen);

      retr->VertexList[VertexNo * VECSIZE + 0] = VD[0] + UD[0] + LIMBS(i)[0];
      retr->VertexList[VertexNo * VECSIZE + 1] = VD[1] + UD[1] + LIMBS(i)[1];
      retr->VertexList[VertexNo * VECSIZE + 2] = VD[2] + UD[2] + LIMBS(i)[2];

      SkinNo++;
    }
  }
}
